import json
from tqdm import tqdm
from time import sleep
from datetime import datetime, timedelta


def get_lb_hourly_costs(session):
    client = session.client("pricing", region_name="us-east-1")

    net_params = {
        "ServiceCode": "AmazonEC2",
        "Filters": [
            {
                "Type": "TERM_MATCH",
                "Field": "productFamily",
                "Value": "Load Balancer-Network",
            }
        ],
    }

    app_params = {
        "ServiceCode": "AmazonEC2",
        "Filters": [
            {
                "Type": "TERM_MATCH",
                "Field": "productFamily",
                "Value": "Load Balancer-Application",
            }
        ],
    }

    # classic_params = {
    #     "ServiceCode": "AmazonEC2",
    #     "Filters": [{"Type": "TERM_MATCH", "Field": "productFamily", "Value": "Load Balancer"}],
    # }

    net_response = client.get_products(**net_params)
    app_response = client.get_products(**app_params)
    # classic_response = client.get_products(**classic_params)

    hourly_costs = {"network": {}, "application": {}}  # , "classic": {}}
    responses = {
        "network": net_response,
        "application": app_response,
    }  # , "classic": classic_response}

    for response in responses:
        for product in responses[response]["PriceList"]:
            data = json.loads(product)

            product_key = list(data["terms"]["OnDemand"].keys())[0]
            price_dimensions_key = list(
                data["terms"]["OnDemand"][product_key]["priceDimensions"].keys()
            )[0]

            if (
                data["terms"]["OnDemand"][product_key]["priceDimensions"][
                    price_dimensions_key
                ]["unit"]
                != "Hrs"
            ):
                continue

            region = data["product"]["attributes"]["regionCode"]
            ppu = data["terms"]["OnDemand"][product_key]["priceDimensions"][
                price_dimensions_key
            ]["pricePerUnit"]

            if region not in hourly_costs[response]:
                hourly_costs[response][region] = ppu

    return hourly_costs


def delete_tgs(session, tgs, dry_run=False):
    elb_client = session.client("elbv2")

    pbar = tqdm(tgs)
    for tg_arn in pbar:
        if not dry_run:
            try:
                elb_client.delete_target_group(TargetGroupArn=tg_arn)
            except Exception as e:
                pbar.write(f"Failed to delete target group {tg_arn} with error {e}")
        pbar.write(f"deleted target group {tg_arn} (dry run: {dry_run})")

    return None


def disable_lb_deletion_protection(client, lb_arn):
    client.modify_load_balancer_attributes(
        LoadBalancerArn=lb_arn,
        Attributes=[{"Key": "deletion_protection.enabled", "Value": "false"}],
    )


def delete_lbs(session, lbs, dry_run=False):
    elb_client = session.client("elbv2")
    waiter = elb_client.get_waiter("load_balancers_deleted")

    pbar = tqdm(lbs)
    for lb_arn in pbar:
        if not dry_run and "populated_target_groups" not in lbs[lb_arn]:
            try:
                lb_attributes = elb_client.describe_load_balancer_attributes(
                    LoadBalancerArn=lb_arn
                )["Attributes"]

                if any(
                    attr["Key"] == "deletion_protection.enabled"
                    and attr["Value"] == "true"
                    for attr in lb_attributes
                ):
                    pbar.write(
                        f"Load balancer {lb_arn} has deletion protection enabled, disabling..."
                    )
                    disable_lb_deletion_protection(elb_client, lb_arn)

                del_result = elb_client.delete_load_balancer(LoadBalancerArn=lb_arn)

                if del_result["ResponseMetadata"]["HTTPStatusCode"] != 200:
                    pbar.write(f"Failed to delete load balancer {lb_arn}: {del_result}")

            except Exception as e:
                pbar.write(f"Failed to delete load balancer {lb_arn} with error {e}")

            pbar.write(f"waiting for load balancer {lb_arn} to be deleted...")
            waiter.wait(
                LoadBalancerArns=[lb_arn],
                WaiterConfig={"Delay": 15, "MaxAttempts": 100},
            )
            sleep(5)
        pbar.write(f"deleted load balancer {lb_arn} (dry run: {dry_run})")

        delete_tgs(session, lbs[lb_arn]["empty_target_groups"])

    return None


def delete_ebs_volumes(volume_ids, session, dry_run=False):
    ec2 = session.resource("ec2")
    for volume_id in volume_ids:
        volume = ec2.Volume(volume_id)
        print("Deleting volume {}".format(volume_id))
        if not dry_run:
            volume.delete()


def estimate_snapshots_cost(session, snapshot_ids):
    ec2 = session.client("ec2")

    # Pricing details (as of September 2021)
    # This value might change, so you should update it based on the current pricing details
    price_per_gb_month = 0.05

    total_size_gb = 0
    for snapshot_id in snapshot_ids:
        snapshot = ec2.describe_snapshots(SnapshotIds=[snapshot_id])["Snapshots"][0]
        total_size_gb += snapshot["VolumeSize"]

    cost = total_size_gb * price_per_gb_month

    return cost


def delete_ebs_snapshots(snapshot_ids, session, dry_run=False):
    ec2 = session.client("ec2")
    for snapshot_id in snapshot_ids:
        print("Deleting snapshot {}".format(snapshot_id))
        if not dry_run:
            ec2.delete_snapshot(SnapshotId=snapshot_id)


def get_old_snapshots(session, days):
    ec2 = session.client("ec2")

    now = datetime.now()
    snapshots = ec2.describe_snapshots(OwnerIds=["self"])["Snapshots"]
    old_snapshots = [
        snapshot["SnapshotId"]
        for snapshot in snapshots
        if (now - snapshot["StartTime"].replace(tzinfo=None)) > timedelta(days=days)
    ]

    return old_snapshots


def scan_for_tgs_no_targets_or_lb(session):
    elb_client = session.client("elbv2")

    unused_target_groups = {
        tg["TargetGroupArn"]: {
            "TargetHealthDescriptions": elb_client.describe_target_health(
                TargetGroupArn=tg["TargetGroupArn"]
            )["TargetHealthDescriptions"],
            "LoadBalancerArns": tg["LoadBalancerArns"],
        }
        for tg in elb_client.describe_target_groups()["TargetGroups"]
    }

    tgs = []

    print("getting target groups with zero targets or no configured load balancer...")
    for target_group_arn in tqdm(unused_target_groups):
        sleep(0.05)

        if len(unused_target_groups[target_group_arn]["TargetHealthDescriptions"]) == 0:
            tgs.append(target_group_arn)

        if len(unused_target_groups[target_group_arn]["LoadBalancerArns"]) == 0:
            tgs.append(target_group_arn)

    return tgs


def scan_for_lbs_no_targets(session, region, omit_pricing=False):
    elb_client = session.client("elbv2")

    print("getting hourly costs of load balancers...")

    if not omit_pricing:
        hourly_costs = get_lb_hourly_costs(session)

    response = elb_client.describe_load_balancers()

    if len(response["LoadBalancers"]) == 0:
        print(f"No load balancers found in region {region}")
        return {"total_monthly_cost": 0}

    lbs = {}

    for lb in tqdm(response["LoadBalancers"]):
        lb_arn = lb["LoadBalancerArn"]

        if not omit_pricing:
            lb_cost_value = (
                float(list(hourly_costs[lb["Type"]][region].values())[0]) * 730
            )
        else:
            lb_cost_value = 0

        lb_target_groups = {
            tg["TargetGroupArn"]: elb_client.describe_target_health(
                TargetGroupArn=tg["TargetGroupArn"]
            )
            for tg in elb_client.describe_target_groups(LoadBalancerArn=lb_arn)[
                "TargetGroups"
            ]
        }

        for lb_target_group_arn in lb_target_groups:
            if (
                len(lb_target_groups[lb_target_group_arn]["TargetHealthDescriptions"])
                == 0
            ):
                if lb_arn not in lbs:
                    lbs[lb_arn] = {
                        "monthly_cost": lb_cost_value,
                        "empty_target_groups": [lb_target_group_arn],
                    }
                else:
                    lbs[lb_arn]["empty_target_groups"].append(lb_target_group_arn)

        if lb_arn in lbs:
            for lb_target_group_arn in lb_target_groups:
                if (
                    len(
                        lb_target_groups[lb_target_group_arn][
                            "TargetHealthDescriptions"
                        ]
                    )
                    != 0
                ):
                    if "populated_target_groups" not in lbs[lb_arn]:
                        lbs[lb_arn]["populated_target_groups"] = [lb_target_group_arn]
                    else:
                        lbs[lb_arn]["populated_target_groups"].append(
                            lb_target_group_arn
                        )

    lbs["total_monthly_cost"] = sum([lbs[lb]["monthly_cost"] for lb in lbs])

    return lbs


def scan_for_unused_ebs_volumes(session):
    client = session.client("ec2")

    cost_per_gb_map = {
        "gp3": 0.08,
        "gp2": 0.1,
        "io1": 0.125,
        "io2": 0.125,
        "st1": 0.045,
        "sc1": 0.025,
        "standard": 0.05,
    }

    volumes = client.describe_volumes()["Volumes"]

    print("getting unused ebs volumes...")
    unused_volumes = {
        "volumes": [
            {
                "VolumeId": volume["VolumeId"],
                "Size": volume["Size"],
                "CreateTime": str(volume["CreateTime"]),
                "MultiAttachEnabled": volume["MultiAttachEnabled"],
                "Attachments": volume["Attachments"],
                "MonthlyCost": volume["Size"] * cost_per_gb_map[volume["VolumeType"]],
            }
            for volume in tqdm(volumes)
            if volume["State"] == "available"
        ]
    }

    unused_volumes["total_monthly_cost"] = sum(
        [volume["MonthlyCost"] for volume in unused_volumes["volumes"]]
    )

    return unused_volumes



from datetime import datetime, timedelta, UTC
from tqdm import tqdm


def get_buckets(session, days):
    cutoff_time = datetime.now(UTC) - timedelta(days=days)

    s3 = session.client("s3")

    s3_buckets = {"old": [], "empty": []}

    response = s3.list_buckets()
    buckets = response["Buckets"]

    pbar = tqdm(buckets)
    for bucket in pbar:
        bucket_name = bucket["Name"]

        response = s3.list_objects_v2(Bucket=bucket_name)
        if "Contents" not in response:
            s3_buckets["empty"].append(bucket_name)
            continue

        newest_object = response["Contents"][0]
        newest_object_time = newest_object["LastModified"]

        if newest_object_time < cutoff_time:
            s3_buckets["old"].append(bucket_name)

    return s3_buckets


def get_bucket_cost(session, bucket_name):
    s3 = session.client("s3")

    size = 0
    paginator = s3.get_paginator("list_objects_v2")
    for page in paginator.paginate(Bucket=bucket_name):
        if "Contents" not in page:
            # bucket is empty
            continue

        for obj in page["Contents"]:
            size += obj["Size"]

    # Calculate the cost based on the size of the data
    # (Assumes the standard S3 pricing model with a cost of $0.023 per GB per month)
    cost = (size / 1024**3) * 0.023

    return cost



from boto3 import Session
from click import group, option, pass_context

import json
from .ec2 import (
    scan_for_lbs_no_targets,
    delete_lbs,
    scan_for_tgs_no_targets_or_lb,
    delete_tgs,
    scan_for_unused_ebs_volumes,
    delete_ebs_volumes,
    get_old_snapshots,
    estimate_snapshots_cost,
)
from .s3 import get_buckets, get_bucket_cost
from .iam import get_unused_iam_roles


@group()
@option("--profile", "-p", required=False, help="AWS profile")
@option("--region", "-r", required=False, help="AWS region")
@pass_context
def cli(ctx, profile, region):
    print("Welcome to the AWS Cost Mutilator!")
    ctx.obj = {}

    if profile and region:
        ctx.obj["session"] = Session(region_name=region, profile_name=profile)
        ctx.obj["profile"] = profile
        ctx.obj["region"] = region
    elif profile:
        ctx.obj["session"] = Session(profile_name=profile)
        ctx.obj["profile"] = profile
        ctx.obj["region"] = ctx.obj["session"].region_name
    elif region:
        ctx.obj["session"] = Session(region_name=region)
        ctx.obj["profile"] = ctx.obj["session"].profile_name
        ctx.obj["region"] = region
    else:
        ctx.obj["session"] = Session()
        ctx.obj["profile"] = ctx.obj["session"].profile_name
        ctx.obj["region"] = ctx.obj["session"].region_name


@cli.group()
@pass_context
def check(ctx):
    pass


@cli.group()
@option("--dry-run", "-d", is_flag=True, help="Perform a dry run")
@pass_context
def clean(ctx, dry_run):
    ctx.obj["dry_run"] = dry_run
    pass


@check.command("s3")
@option(
    "--days",
    type=int,
    default=365,
    help="Find empty buckets and buckets with no objects newer than this number of days",
)
@pass_context
def s3_(ctx, days):
    session = ctx.obj["session"]
    profile = ctx.obj["profile"]
    buckets = get_buckets(session, days)
    cost = 0
    for bucket_name in buckets["old"]:
        cost += get_bucket_cost(session, bucket_name)
    print(json.dumps(buckets, indent=4))
    print(
        f"Run:\n\nacm --profile {profile} clean s3 --days {days}\n\nto delete these resources and save ${cost:.2f} per month"
    )


@check.command("roles")
@option("--days", type=int, help="Find roles unused for this many days")
@pass_context
def roles_(ctx, days):
    session = ctx.obj["session"]
    unused_roles = get_unused_iam_roles(session, days)

    if len(unused_roles) == 0:
        print("No unused IAM roles found!")
        exit(0)

    print(
        f"There are {len(unused_roles)} IAM roles unused for more than {days} {'day' if days == 1 else 'days'}:"
    )
    print(json.dumps(unused_roles, indent=4))


@check.command("ebs")
@pass_context
def ebs_(ctx):
    session = ctx.obj["session"]
    region = ctx.obj["region"]
    profile = ctx.obj["profile"]
    unused_ebs_volumes = scan_for_unused_ebs_volumes(session)
    total_monthly_cost = unused_ebs_volumes["total_monthly_cost"]
    del unused_ebs_volumes["total_monthly_cost"]

    if len(unused_ebs_volumes) == 0:
        print("No unused EBS volumes found!")
        return

    print(f"There are {len(unused_ebs_volumes['volumes'])} unused EBS volumes:")
    print(json.dumps(unused_ebs_volumes["volumes"], indent=4))
    print(
        f"Run:\n\nacm clean ebs --region {region} --profile {profile}\n\nto delete these resources and save ${total_monthly_cost:.2f} per month"
    )
    exit(0)


@check.command("ebssnap")
@option("--older-than", type=int, help="Find snapshots older than this many days")
@pass_context
def ebs_snapshots_(ctx, older_than):
    session = ctx.obj["session"]
    region = ctx.obj["region"]
    profile = ctx.obj["profile"]
    old_snapshots = get_old_snapshots(session, older_than)

    if len(old_snapshots) == 0:
        print("No old EBS snapshots found!")
        exit(0)

    else:
        total_monthly_cost = estimate_snapshots_cost(session, old_snapshots)

    print(
        f"There are {len(old_snapshots)} EBS snapshots older than {older_than} {'day' if older_than == 1 else 'days'}:"
    )
    print(json.dumps(old_snapshots, indent=4))
    print(
        f"Run:\n\nacm clean ebsnap --region {region} --profile {profile} --older-than {older_than}\n\nto delete these resources and save ${total_monthly_cost:.2f} per month"
    )
    exit(0)


@check.command("tgs")
@pass_context
def tgs_(ctx):
    session = ctx.obj["session"]
    target_groups = scan_for_tgs_no_targets_or_lb(session)

    if len(target_groups) == 0:
        print("No target groups without targets or load balancers found!")
        return

    print(
        f"There are {len(target_groups)} target groups with zero targets or no load balancer:"
    )
    print(json.dumps(target_groups, indent=4))


@check.command("lbs")
@pass_context
def lbs_(ctx):
    # Perform analysis of ELBv2 resources in the specified region and profile
    session = ctx.obj["session"]
    region = ctx.obj["region"]
    profile = ctx.obj["profile"]
    load_balancers = scan_for_lbs_no_targets(session, region, region)

    total_monthly_cost = load_balancers["total_monthly_cost"]
    del load_balancers["total_monthly_cost"]
    num_lbs_no_targets = len(load_balancers)

    if num_lbs_no_targets == 0:
        print("No load balancers without targets found!")
        return

    print(f"There are {len(load_balancers)} load balancers with empty target groups:")
    print(json.dumps(load_balancers, indent=4))
    print(
        f"Run:\n\nacm clean elbv2 --region {region} --profile {profile}\n\nto delete these resources and save ${total_monthly_cost:.2f} per month"
    )

    exit(0)


# CLEAN COMMANDS


@clean.command("s3")
@option(
    "--days",
    type=int,
    default=365,
    help="Delete empty buckets and buckets with no objects newer than this number of days",
)
@pass_context
def s3(ctx, days):
    raise NotImplementedError("This feature is not yet implemented")


@clean.command("tgs")
@pass_context
def tgs(ctx):
    session = ctx.obj["session"]
    dry_run = ctx.obj["dry_run"]
    target_groups = scan_for_tgs_no_targets_or_lb(session)

    num_tgs = len(target_groups)

    if num_tgs == 0:
        print("No target groups without targets or load balancers found!")
        return

    print(
        f"There are {len(target_groups)} target groups with zero targets or no load balancer:"
    )
    print(json.dumps(target_groups, indent=4))

    # Ask the user for confirmation
    response = input(
        f"Are you sure you want to continue? This will delete {num_tgs} target groups. (yes/no): "
    )

    # Check the user's response
    if response == "yes":
        # Execute the code if the response was "yes"
        if dry_run:
            print("Dry run mode enabled, no resources will be deleted.")

        delete_tgs(session, target_groups, dry_run)

        print(f"Deleted {num_tgs} target groups.")
        # print(
        #     f"Deleted {num_tgs} load balancers and their associated target groups saving ${total_monthly_cost:.2f} per month."
        # )

    else:
        # Exit the program if the response was "no" or anything else
        print("Aborted")

    exit(0)


@clean.command("lbs")
@pass_context
def lbs(ctx):
    # Perform analysis of ELBv2 resources in the specified region and profile
    session = ctx.obj["session"]
    region = ctx.obj["region"]
    dry_run = ctx.obj["dry_run"]
    load_balancers = scan_for_lbs_no_targets(session, region)
    total_monthly_cost = load_balancers["total_monthly_cost"]
    del load_balancers["total_monthly_cost"]
    num_lbs = len(load_balancers)

    if num_lbs == 0:
        print("No load balancers without targets found!")
        return

    print(f"There are {num_lbs} load balancers with empty target groups:")
    print(json.dumps(load_balancers, indent=4))

    # Ask the user for confirmation
    response = input(
        f"Are you sure you want to continue? This will delete {num_lbs} load balancers and their associated target groups. If there are both populated and empty target groups associated with a load balancer, it will detach and delete only the empty target groups, leaving the load balancer and populated target groups in place. (yes/no): "
    )

    # Check the user's response
    if response == "yes":
        # Execute the code if the response was "yes"
        if dry_run:
            print("Dry run mode enabled, no resources will be deleted.")

        delete_lbs(session, load_balancers, dry_run)

        print(
            f"Deleted {num_lbs} load balancers and their associated target groups saving ${total_monthly_cost:.2f} per month."
        )

    else:
        # Exit the program if the response was "no" or anything else
        print("Aborted")

    exit(0)


@clean.command("ebs")
@pass_context
def ebs(ctx):
    session = ctx.obj["session"]
    dry_run = ctx.obj["dry_run"]
    unused_ebs_volumes = scan_for_unused_ebs_volumes(session)
    total_monthly_cost = unused_ebs_volumes["total_monthly_cost"]
    del unused_ebs_volumes["total_monthly_cost"]

    if len(unused_ebs_volumes) == 0:
        print("No unused EBS volumes found!")
        return

    print(f"There are {len(unused_ebs_volumes['volumes'])} unused EBS volumes:")
    print(json.dumps(unused_ebs_volumes["volumes"], indent=4))

    # Ask the user for confirmation
    response = input(
        f"Are you sure you want to continue? This will delete {len(unused_ebs_volumes['volumes'])} EBS volumes. (yes/no): "
    )
    if response == "yes":
        # Execute the code if the response was "yes"
        if dry_run:
            print("Dry run mode enabled, no resources will be deleted.")

        delete_ebs_volumes(
            session, [vol["VolumeId"] for vol in unused_ebs_volumes], dry_run
        )
        print(
            f"Deleted {len(unused_ebs_volumes['volumes'])} EBS volumes saving ${total_monthly_cost:.2f} per month."
        )
    else:
        # Exit the program if the response was "no" or anything else
        print("Aborted")

    exit(0)


if __name__ == "__main__":
    cli()



from datetime import datetime, timedelta, timezone


def get_unused_iam_roles(session, days):
    iam = session.client("iam")
    cloudtrail = session.client("cloudtrail")
    unused_roles = []

    paginator = iam.get_paginator("list_roles")
    for response in paginator.paginate():
        for role in response["Roles"]:
            if role["RoleName"] == "cryptoassets-solana-dev-response-handler":
                pass
            if role["Path"].startswith("/aws-service-role/"):
                continue

            events = cloudtrail.lookup_events(
                LookupAttributes=[
                    {"AttributeKey": "ResourceName", "AttributeValue": role["RoleName"]}
                ],
                MaxResults=1,
            )

            # If the role has no events in CloudTrail, consider it unused
            if not events["Events"]:
                unused_roles.append(role["RoleName"])
                continue

            # Get the time of the last event
            last_event_time = events["Events"][0]["EventTime"]

            # If the last event is older than the specified number of days, consider the role unused
            if last_event_time < datetime.now(timezone.utc) - timedelta(days=days):
                unused_roles.append(role["RoleName"])

    return unused_roles



from .ec2 import delete_tgs, delete_lbs, disable_lb_deletion_protection
from .ec2 import (
    scan_for_tgs_no_targets_or_lb,
    scan_for_lbs_no_targets,
    delete_ebs_volumes,
)
from .s3 import get_buckets, get_bucket_cost
import boto3
from moto import mock_elbv2, mock_ec2, mock_s3


# Create a mock Elastic Load Balancing client
@mock_elbv2
def test_delete_tgs():
    # Create a mock Elastic Load Balancing client
    elb_client = boto3.client("elbv2")

    # Define the expected response of the mock describe_target_groups method
    expected_response = {"TargetGroups": []}

    # Create a mock describe_target_groups method that returns the expected response
    elb_client.describe_target_groups = lambda **kwargs: expected_response

    # Test deleting target groups
    session = boto3.Session(region_name="us-west-2", profile_name="default")
    tgs = ["tg-1234567890", "tg-0987654321"]
    dry_run = False
    delete_tgs(session, tgs, dry_run)

    # Verify that the target groups were deleted
    response = elb_client.describe_target_groups()
    assert response == expected_response

    # Test dry run deleting target groups
    session = boto3.Session(region_name="us-west-2", profile_name="default")
    tgs = ["tg-1234567890", "tg-0987654321"]
    dry_run = True
    delete_tgs(session, tgs, dry_run)

    # Verify that the target groups were not deleted
    response = elb_client.describe_target_groups(TargetGroupArns=tgs)
    assert response == expected_response


@mock_ec2
@mock_elbv2
def test_delete_lbs():
    # Create a mock AWS session
    session = boto3.Session(
        region_name="us-east-1",
        aws_access_key_id="mock_access_key",
        aws_secret_access_key="mock_secret_key",
    )

    # Create a mock VPC and a mock subnet
    ec2_client = session.client("ec2")
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(
        VpcId=vpc_id,
        CidrBlock="10.0.0.0/24",
    )[
        "Subnet"
    ]["SubnetId"]

    # Create a mock Elastic Load Balancer
    elb_client = session.client("elbv2")
    elb_response = elb_client.create_load_balancer(
        Name="mock-elb",
        Subnets=[subnet_id],
        SecurityGroups=["sg-12345678"],
    )
    elb_arn = elb_response["LoadBalancers"][0]["LoadBalancerArn"]

    # Create a mock target group
    tg_response = elb_client.create_target_group(
        Name="mock-tg",
        Protocol="HTTP",
        Port=80,
        VpcId=vpc_id,
    )
    tg_arn = tg_response["TargetGroups"][0]["TargetGroupArn"]

    # Add the target group to the load balancer
    elb_client.create_listener(
        LoadBalancerArn=elb_arn,
        Protocol="HTTP",
        Port=80,
        DefaultActions=[
            {
                "Type": "forward",
                "TargetGroupArn": tg_arn,
            },
        ],
    )

    # Create the input argument for the delete_lbs function
    lbs = {
        elb_arn: {
            "empty_target_groups": [tg_arn],
        }
    }

    # Test the delete_lbs function with dry_run=True
    delete_lbs(session, lbs, dry_run=True)

    # Verify that the load balancer and target group were not deleted
    elb_response = elb_client.describe_load_balancers()
    assert len(elb_response["LoadBalancers"]) == 1

    tg_response = elb_client.describe_target_groups()
    assert len(tg_response["TargetGroups"]) == 1

    # Test the delete_lbs function with dry_run=False
    delete_lbs(session, lbs)

    # Verify that the load balancer and target group were deleted
    elb_response = elb_client.describe_load_balancers()
    assert len(elb_response["LoadBalancers"]) == 0

    tg_response = elb_client.describe_target_groups()
    assert len(tg_response["TargetGroups"]) == 0


@mock_ec2
@mock_elbv2
def test_disable_lb_deletion_protection():
    # Create a mock AWS session
    session = boto3.Session(
        region_name="us-east-1",
        aws_access_key_id="mock_access_key",
        aws_secret_access_key="mock_secret_key",
    )

    # Create a mock VPC and a mock subnet
    ec2_client = session.client("ec2")
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(
        VpcId=vpc_id,
        CidrBlock="10.0.0.0/24",
    )[
        "Subnet"
    ]["SubnetId"]

    # Create a mock Elastic Load Balancer
    elb_client = session.client("elbv2")
    elb_response = elb_client.create_load_balancer(
        Name="mock-elb",
        Subnets=[subnet_id],
        SecurityGroups=["sg-12345678"],
    )
    elb_arn = elb_response["LoadBalancers"][0]["LoadBalancerArn"]

    # Enable deletion protection for the load balancer
    elb_client.modify_load_balancer_attributes(
        LoadBalancerArn=elb_arn,
        Attributes=[{"Key": "deletion_protection.enabled", "Value": "true"}],
    )

    # Call the disable_lb_deletion_protection function
    disable_lb_deletion_protection(elb_client, elb_arn)

    # Verify that deletion protection has been disabled
    elb_response = elb_client.describe_load_balancer_attributes(LoadBalancerArn=elb_arn)
    attributes = elb_response["Attributes"]
    assert any(
        attr["Key"] == "deletion_protection.enabled" and attr["Value"] == "false"
        for attr in attributes
    )


@mock_ec2
@mock_elbv2
def test_scan_for_tgs_no_targets_or_lb():
    # Create a mock AWS session
    session = boto3.Session(
        region_name="us-east-1",
        aws_access_key_id="mock_access_key",
        aws_secret_access_key="mock_secret_key",
    )

    # Create a mock VPC and a mock subnet
    ec2_client = session.client("ec2")
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(
        VpcId=vpc_id,
        CidrBlock="10.0.0.0/24",
    )[
        "Subnet"
    ]["SubnetId"]

    # Create a mock Elastic Load Balancer
    elb_client = session.client("elbv2")
    elb_response = elb_client.create_load_balancer(
        Name="mock-elb",
        Subnets=[subnet_id],
        SecurityGroups=["sg-12345678"],
    )
    elb_arn = elb_response["LoadBalancers"][0]["LoadBalancerArn"]

    # Create a mock target group
    tg_response = elb_client.create_target_group(
        Name="mock-tg",
        Protocol="HTTP",
        Port=80,
        VpcId=vpc_id,
    )
    tg_arn = tg_response["TargetGroups"][0]["TargetGroupArn"]

    # Add the target group to the load balancer
    listener_arn = elb_client.create_listener(
        LoadBalancerArn=elb_arn,
        Protocol="HTTP",
        Port=80,
        DefaultActions=[
            {
                "Type": "forward",
                "TargetGroupArn": tg_arn,
            },
        ],
    )["Listeners"][0]["ListenerArn"]

    # Call the get_tgs_no_targets_or_lb function
    target_groups = scan_for_tgs_no_targets_or_lb(session)

    # Verify that the target group is returned by the function
    assert tg_arn in target_groups

    ami_id = ec2_client.describe_images()["Images"][0]["ImageId"]

    # Create a mock EC2 instance
    ec2_response = ec2_client.run_instances(
        ImageId=ami_id,
        InstanceType="t2.micro",
        MaxCount=1,
        MinCount=1,
        NetworkInterfaces=[
            {
                "SubnetId": subnet_id,
                "DeviceIndex": 0,
                "AssociatePublicIpAddress": True,
            },
        ],
    )

    # Get Instance Id
    instance_id = ec2_response["Instances"][0]["InstanceId"]

    # Add a target to the target group
    elb_client.register_targets(
        TargetGroupArn=tg_arn,
        Targets=[
            {
                "Id": instance_id,
                "Port": 80,
            },
        ],
    )

    # Call the get_tgs_no_targets_or_lb function
    target_groups = scan_for_tgs_no_targets_or_lb(session)

    # Verify that the target group is not returned by the function
    assert tg_arn not in target_groups

    # Remove the load balancer from the target group
    elb_client.delete_listener(
        ListenerArn=listener_arn,
    )
    elb_client.deregister_targets(
        TargetGroupArn=tg_arn,
        Targets=[
            {
                "Id": instance_id,
                "Port": 80,
            },
        ],
    )

    # Call the get_tgs_no_targets_or_lb function
    target_groups = scan_for_tgs_no_targets_or_lb(session)

    # Verify that the target group is returned by the function
    assert tg_arn in target_groups


@mock_ec2
@mock_elbv2
def test_scan_for_lbs_no_targets():
    region = "us-east-1"
    # Create a mock AWS session
    session = boto3.Session(
        region_name="us-east-1",
        aws_access_key_id="mock_access_key",
        aws_secret_access_key="mock_secret_key",
    )

    # Create a mock VPC and a mock subnet
    ec2_client = session.client("ec2")
    vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
    subnet_id = ec2_client.create_subnet(
        VpcId=vpc_id,
        CidrBlock="10.0.0.0/24",
    )[
        "Subnet"
    ]["SubnetId"]

    # Create a mock Elastic Load Balancer
    elb_client = session.client("elbv2")
    elb_response = elb_client.create_load_balancer(
        Name="mock-elb",
        Subnets=[subnet_id],
        SecurityGroups=["sg-12345678"],
    )
    elb_arn = elb_response["LoadBalancers"][0]["LoadBalancerArn"]

    # Create a mock target group
    tg_response = elb_client.create_target_group(
        Name="mock-tg",
        Protocol="HTTP",
        Port=80,
        VpcId=vpc_id,
    )
    tg_arn = tg_response["TargetGroups"][0]["TargetGroupArn"]

    # Add the target group to the load balancer
    elb_client.create_listener(
        LoadBalancerArn=elb_arn,
        Protocol="HTTP",
        Port=80,
        DefaultActions=[
            {
                "Type": "forward",
                "TargetGroupArn": tg_arn,
            },
        ],
    )["Listeners"][0]["ListenerArn"]

    # Call the get_lbs_no_targets function
    load_balancers = scan_for_lbs_no_targets(session, region, omit_pricing=True)

    assert elb_arn in load_balancers


@mock_ec2
def test_scan_for_unused_ebs_volumes():
    session = boto3.Session(region_name="us-east-1")
    client = session.client("ec2")

    # Create an EBS volume
    client.create_volume(AvailabilityZone="us-east-1a", Size=1, VolumeType="gp2")

    # Get the list of all EBS volumes in the region
    volumes = client.describe_volumes()["Volumes"]

    # Filter the list to include only volumes that are not attached to any EC2 instances
    unused_volumes = [volume for volume in volumes if volume["State"] == "available"]

    # Assert that the list is not empty
    assert len(unused_volumes) > 0

    # Assert that the list contains the EBS volume that we created
    assert len([volume for volume in unused_volumes if volume["Size"] == 1]) == 1


@mock_ec2
def test_delete_ebs_volumes():
    session = boto3.Session(region_name="us-east-1")
    ec2 = session.resource("ec2")

    # create a volume to delete
    volume = ec2.create_volume(AvailabilityZone="us-east-1a", Size=10)

    # delete the volume and confirm it is deleted
    assert delete_ebs_volumes([volume.id], session) is None
    assert list(ec2.volumes.all()) == []


@mock_s3
def test_get_old_buckets():
    # Create a boto3 session for testing
    session = boto3.Session(region_name="us-east-1")

    # Create a client for the S3 service
    s3 = session.client("s3")

    # Create a test bucket
    s3.create_bucket(Bucket="test-bucket")

    # Check that the function correctly identifies an empty bucket as old
    buckets = get_buckets(session, 365)
    assert "test-bucket" in buckets["empty"]

    # Add a new object to the test bucket
    s3.put_object(Bucket="test-bucket", Key="test.txt", Body=b"test")

    # Check that the function does not consider the bucket to be old
    buckets = get_buckets(session, 365)
    assert "test-bucket" not in buckets["old"]


@mock_s3
def test_get_bucket_cost():
    # Create a boto3 session for testing
    session = boto3.Session(region_name="us-east-1")

    # Create a client for the S3 service
    s3 = session.client("s3")

    # Create a test bucket
    s3.create_bucket(Bucket="test-bucket")

    # Check that the cost of an empty bucket is 0
    cost = get_bucket_cost(session, "test-bucket")
    assert cost == 0

    # Add an object to the test bucket
    s3.put_object(Bucket="test-bucket", Key="test.txt", Body=b"0123456789")

    # Check that the cost of the bucket is calculated correctly
    cost = get_bucket_cost(session, "test-bucket")
    assert cost > 0
